use crossbeam::channel::{Receiver, Sender};
use rand::Rng;
use std::sync::mpsc;

#[derive(Clone)]
/// A generic job dispatcher struct that allows sending jobs of type `I` and receiving results of type `O` using message passing.
pub struct JobDispatcher<I: Send + 'static, O: Send + 'static> {
    sender: Sender<JobOrder<I, O>>,
}

pub struct JobResult<O>(std::sync::mpsc::Receiver<O>);

impl<O> JobResult<O> {
    /// Wait for the Result of a Job.
    pub fn wait(self) -> O {
        self.0.recv().unwrap()
    }
}

impl<I: Send + 'static, O: Send + 'static> JobDispatcher<I, O> {
    /// Creates a new instance of `JobDispatcher` and returns a tuple that contains it and a receiver end for `JobOrder`s.
    /// # Example:
    /// ```
    /// use jobdispatcher::*;
    /// // Create job dispatcher
    /// let (dispatcher, recv) = JobDispatcher::<i32, i32>::new();
    ///
    /// // Worker Thread
    /// std::thread::spawn(move || {
    ///     for job in recv {
    ///         let result = job.param + 1;
    ///         job.done(result);
    ///     }
    /// });
    ///    
    /// // Usage
    /// let result = dispatcher.send(3);
    /// assert_eq!(result, 4);
    /// ```
    #[must_use]
    pub fn new() -> (Self, Receiver<JobOrder<I, O>>) {
        let (sender, receiver) = crossbeam::channel::bounded(12);

        (Self { sender: sender }, receiver)
    }

    /// Sends a job of type `T` to the job dispatcher and waits for its result of type `V`.
    /// Returns the result of the job once it has been processed.
    /// # Panics
    /// This function panics when the `JobOrder` struct gets out of scope without returning a finished result.
    /// Additionally if the internal `Mutex` is poisoned, this function will panic as well.
    pub fn send(&self, param: I) -> O {
        let (tx, rx) = mpsc::channel();
        let job_order = JobOrder::new(param, move |ret| {
            tx.send(ret).unwrap();
        });
        self.sender.send(job_order).unwrap();
        rx.recv().unwrap()
    }

    pub fn send_async(&self, param: I) -> JobResult<O> {
        let (tx, rx) = mpsc::channel();
        let job_order = JobOrder::new(param, move |ret| {
            tx.send(ret).unwrap();
        });
        self.sender.send(job_order).unwrap();
        JobResult(rx)
    }

    /// Sends a job of type `T` to the job dispatcher and waits for its result of type `V`.
    /// Returns `Some(V)` when the job returns an result, `None` if somehow nothing was returned or the internal `Mutex` is poisoned.
    pub fn try_send(&self, param: I) -> Option<O> {
        let (tx, rx) = mpsc::channel();
        let job_order = JobOrder::new(param, move |ret| {
            tx.send(ret).unwrap();
        });
        self.sender.send(job_order).ok()?;
        rx.recv().ok()
    }
}

/// A struct that represents a job order that encapsulates a job of type `I` and its result of type `O`, along with a callback function that will send the result back to the job origin.
pub struct JobOrder<I, O> {
    /// The job parameter of type `T`.
    pub param: I,
    callback: Box<dyn FnOnce(O) + Send>,
}

impl<I, O> JobOrder<I, O> {
    /// Creates a new `JobOrder` instance with the specified job parameter `param` of type `I` and a callback function that takes the job result of type `O` as an argument.
    #[must_use]
    fn new(param: I, callback: impl FnOnce(O) + Send + 'static) -> Self {
        Self {
            param,
            callback: Box::new(callback),
        }
    }

    /// Send the result of the `JobOrder` back to it's origin
    pub fn done(self, val: O) {
        (self.callback)(val);
    }
}

pub struct JobMultiplexer<I: Send + 'static, O: Send + 'static> {
    dispatchers: Vec<JobDispatcher<I, O>>,
}

fn get_random_item<T>(list: &[T]) -> Option<&T> {
    if list.is_empty() {
        return None;
    }
    let mut rng = rand::rng();
    let index = rng.random_range(0..list.len());
    list.get(index)
}

impl<I: Send + 'static, O: Send + 'static> JobMultiplexer<I, O> {
    pub fn from(dispatchers: Vec<JobDispatcher<I, O>>) -> Self {
        Self { dispatchers }
    }

    pub fn send(&self, param: I) -> O {
        let d = get_random_item(&self.dispatchers).unwrap();
        d.send(param)
    }

    pub fn send_async(&self, param: I) -> JobResult<O> {
        let d = get_random_item(&self.dispatchers).unwrap();
        d.send_async(param)
    }
}

impl<I: Clone + Send + 'static, O: Send + 'static> JobMultiplexer<I, O> {
    pub fn send_all(&self, param: I) {
        for d in &self.dispatchers {
            let _ = d.send(param.clone());
        }
    }
}
